/*
 * Header:
 * File: libsclient.h
 * Author: Noda Itsuki
 * Version: 7.00
 * Modified by : Ehsan Foroughi, Hamed Hatami ( Sharif University of Tech. )
 * Compatible with server 7.00 ( Connect with version 6 )
 * Date: 2001/04/1
 * EndHeader:
 */

#ifndef ___LIBSCLIENT_H_
#define ___LIBSCLIENT_H_

#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <string.h>

/* ======================================================================
 * 
 */

/* --------------------------------------------------
 * Bool Type
 */ 
/*
typedef enum {
  FALSE = 0,
  TRUE = 1,
} Bool ;
*/
#define Bool bool
#define TRUE true
#define FALSE false

#ifndef M_PI /* in solaris */
#define M_PI (3.1415927)
#endif

/* --------------------------------------------------
 * Number of Objects
 */
#define N_PLAYER_PER_TEAM 	11 
#define N_PLAYER 		 (N_PLAYER_PER_TEAM * 2) 
#define N_LINE 			4 
#define N_GOAL 			2 
#define N_FLAG 			(6 + 6 + 4 + 22 + 14 + 1) 
	/* corner, penalty, goal post, touchline, goalline, center mark */

#define SEENOBJMAX		(N_PLAYER + N_LINE + N_GOAL + N_FLAG)

/* --------------------------------------------------
 * Field Param
 */
#define PITCH_LENGTH 		105.0 
#define PITCH_WIDTH 		68.0 
#define PENALTY_AREA_LENGTH 	16.5
#define PENALTY_AREA_WIDTH 	40.32 
#define GOAL_WIDTH		14.02

/* --------------------------------------------------
 * Error Values
 */
#define DIST_ERR		-1 
#define DIR_ERR 		-360 
/* #define CHNG_ERR	1024 */
#define CHNG_ERR 		1024
#define UNUM_ERR		-1

/* --------------------------------------------------
 * Misc & Utility
 */
#define BUFSIZE 		4096
#define NAMEMAX			256
#define COMMAX			16


#define Abs(x)  (((x) > 0.0)?(x):(-(x)))
#define Pow(x)  ((x)*(x))
#define Rad2Deg(x)  ((double)(x) * 180.0 / M_PI)
#define Deg2Rad(x)  ((double)(x) * M_PI / 180.0)

#ifdef MEMSET
#define bzero(a, b) memset(a, 0, b)
#endif

/* ======================================================================
 * Socket 
 */
typedef struct socket{
  int socketfd  ;
  struct sockaddr_in serv_addr ;
} Socket ;

/* ======================================================================
 *  Sensor Infomation
 */
/* --------------------------------------------------
 *  Side
 */
typedef enum _Side {
  S_UNKNOWN,
  S_LEFT,
  S_RIGHT,
} Side ;

/* --------------------------------------------------
 *  PlayMode
 */
#include "playmode.h"

/* --------------------------------------------------
 *  Init
 */
typedef struct _InitInfo {
  Side side ;
  int unum ;
  char playmodestr[32] ;
  PlayMode playmode ;
} InitInfo ;

/* --------------------------------------------------
 *  View Mode
 */
typedef enum _ViewWidth {
  VW_Narrow,
  VW_Normal,
  VW_Wide,
} ViewWidth ;

typedef enum _ViewQuality {
  VQ_high,
  VQ_low,
} ViewQuality ;

/* --------------------------------------------------
 *  Position Info
 */

typedef struct _XYPos {
  double x ;
  double y ;
} Pos ;

typedef struct _PosState {
  Pos pos ;
  double dir ;
} PosState ;


/* --------------------------------------------------
 *  Object Type
 */
typedef enum _ObjectType {
  OT_Unknown,
  OT_Ball,
  OT_Player,
  OT_Goal,
  OT_Flag,
  OT_Line,
} ObjectType ;

/* --------------------------------------------------
 *  Flags
 */
#include "flagtable.h"

/* --------------------------------------------------
 *  Lines
 */
typedef enum _LinePos {
  LP_Unknown,
  LP_L,
  LP_T,
  LP_B,
  LP_R, 
} LinePos ;


/* --------------------------------------------------
 *  PlayerInfo
 */
typedef enum _WhichTeam {
  WT_Opponent = -1,
  WT_Unknown = 0,
  WT_Our = 1,
} WhichTeam ;

typedef struct _Ppos {
  WhichTeam team ;
  int unum ;
} PlayerInfo ;

/* --------------------------------------------------
 *  See Info
 */
typedef struct _SeenObj {
  ObjectType type ;
  Bool inViewConeP ; /* True if the object is in the view cone */
  int begin;
  int end;
  union {
    PlayerInfo player ;
    Side side ;
    FlagPos fpos ;
    LinePos lpos ;
  } id ;
  double dist ;
  double dir ;
  double dist_chng ;
  double dir_chng ;
  double body_dir ;
  double face_dir ;
} SeenObj ;

typedef struct _SeeInfo {
  int time ;
  int n_obj ;
  SeenObj obj[SEENOBJMAX] ;
} SeeInfo ;

/* --------------------------------------------------
 *  Hear Info
 */
typedef enum _SenderType {
  ST_Self,
  ST_Referee,
  ST_Player,
  ST_OLCoach_l,
  ST_OLCoach_r,
} SenderType ;

typedef struct _HearInfo {
  int time ;
  SenderType sender ;
  double direction ;
  char message[BUFSIZE] ;
} HearInfo ;

/* --------------------------------------------------
 *  Sense Body
 */
typedef struct _SenseBodyInfo {
  int time ;
  ViewQuality viewQuality ;
  ViewWidth viewWidth ;
  double stamina ;
  double effort ;
  double speed_val ;
  double speed_dir ;
  double head_angle ;
  int countKick ;
  int countDash ;
  int countTurn ;
  int countSay ;
  int countTurnNeck ;
  int countCatch ;
  int countMove ;
  int countChangeView ;
} SenseBodyInfo ;

/* --------------------------------------------------
 *  Sensor Info
 */

typedef enum _InfoType {
  SensorT,
  ParamT,
  FullStateT,
  SeeGlobalT,
  ErrorT,
} InfoType ;

typedef enum _SensorInfoType {
  SIT_See,
  SIT_Hear,
  SIT_SenseBody,
} SensorInfoType ;

typedef enum _ParamInfoType {
  PIT_ServerParam,
  PIT_PlayerParam,
  PIT_PlayerType,
} ParamInfoType ;

typedef struct _SensorInfo {
  SensorInfoType type ;
  union {
    SeeInfo see ;
    HearInfo hear ;
    SenseBodyInfo senseBody ;
  } info ;
} SensorInfo ;

typedef struct _ParamInfo {
  ParamInfoType type ;
  union {
    char server_param[BUFSIZE];
    char player_param[BUFSIZE];
    char player_type[BUFSIZE];
  } info ;
} ParamInfo ;

typedef char FullStateInfo[BUFSIZE];

typedef char SeeGlobalInfo[BUFSIZE];

typedef char ErrorInfo[BUFSIZE];

typedef struct _ReceivedInfo {
  InfoType infotype ;
  union {
    SensorInfo sensor ;
    ParamInfo param ;
    FullStateInfo fullstate ;
    SeeGlobalInfo see_global ;
    ErrorInfo error_msg ;
  } info ;
} ReceivedInfo ;

/* ======================================================================
 *  Prototypes 
 */

/* --------------------------------------------------
 * netif.c
 */
extern Socket init_connection0(char *host, int port, Bool nonblockp);
extern Socket init_connection(char *host, int port) ;
extern Socket init_blocking_connection(char *host, int port) ;
extern Bool    send_message(Socket sock, char *buf) ;
extern Bool    receive_message(Socket *sock, char *buf, int size) ;
extern void close_connection(Socket sock) ;


/* --------------------------------------------------
 * sensor_info
 */
extern char *next_token(char *buf) ;
extern const FlagInfo Flags[] ;
extern Bool receive_info(Socket* socket, char* buffer, int size) ;
extern Bool is_sense_body(char *buffer);
extern Bool scan_info(char *buffer, ReceivedInfo *sinfo, char *tname) ;
extern Bool scan_visual_info(char *buffer, SeeInfo *sinfo, char *tname) ;
extern Bool scan_auditorial_info(char *buffer, HearInfo *hinfo) ;
extern Bool scan_senseBody_info(char *buffer, SenseBodyInfo *sbinfo) ;
extern Bool scan_info_online_coach(char *buffer, ReceivedInfo *sinfo, char *teamname) ;
extern Bool scan_visual_info_online_coach(char *buffer, SeeInfo *sinfo, char *teamname) ;

/* --------------------------------------------------
 * basic_com
 */
const PlayModeEntry pm_entry[] = PLAYMODE_TABLE ;
extern PlayMode get_playmode(char *pmodestr);
extern InitInfo send_com_init(Socket *sock, char *teamname, 
			      double version, Bool goaliep) ;
extern InitInfo send_com_init_online_coach(Socket *sock, char *teamname, double version) ;
extern InitInfo send_com_reconnect(Socket *sock, char *teamname, int unum) ;
extern Bool send_com_move(Socket *sock, double x, double y) ;
extern Bool send_com_turn(Socket *sock, double moment) ;
extern Bool send_com_dash(Socket *sock, double power) ;
extern Bool send_com_kick(Socket *sock, double power, double dir) ;
extern Bool send_com_say(Socket *sock, char *message) ;
extern Bool send_com_change_view(Socket *sock, ViewWidth width, 
				 ViewQuality quality) ;
extern Bool send_com_sense_body(Socket *sock) ;
extern Bool send_com_catch(Socket *sock, double dir) ;
extern Bool send_com_bye(Socket *sock) ;
extern Bool send_com_turn_neck(Socket *sock, double dir) ;
extern Bool send_com_look(Socket *sock) ;
Bool send_com_eye(Socket *sock, Bool on) ;

/* --------------------------------------------------
 * position.c
 */
extern Bool estimate_current_pos(SeeInfo *sinf, Side side, PosState *pstate);
extern Bool estimate_current_pos2(SeeInfo *sinf, Side side, PosState *pstate);
extern Bool estimate_current_pos3(SeeInfo *sinf, Side side, PosState *pstate);
extern Bool estimate_current_pos4(SeeInfo *sinf, Side side, PosState *pstate, PosState *errstate);
extern Bool estimate_current_pos5(SeeInfo *sinf, Side side, PosState *pstate);

#endif /* ___LIBSCLIENT_H_ */








