/*
  Log it inorder to debug easily: 
    functions for logging anything

  Ehsan Foroughi 
  ParsAI Team
  Sharif University of Tech.

 */

#define _logs_extern_
#include "logs.h"

#include "headers.h"

void report_error(char *msg)
{
  if (!REPORT_ERR) return;
  fprintf(stderr, "\nTime: %d - Panic : ", world.time.clock);
  fprintf(stderr, msg);
}

void get_server_error(char *msg)
{
  if (!REPORT_ERR) return;
  fprintf(stderr, "\nTime: %d - Server Error : %s", world.time.clock, msg);
}

void logmsg(char *msg)
{
  if (!LOGIT) return;
  fprintf(client.log.file, "\nLogging message in time %d : \n  %s\n", world.time.clock, msg);
}

void logrcvdmsg(char *msg)
{
  if ((!LOGIT)||(!LOG_RCV)) return;
  fprintf(client.log.file, "\nMessage received from server in time %d : \n  %s", world.time.clock, msg);
}

void logplayerstate(void)
{
  if ((!LOGIT)||(!LOG_PLAYER_ST)) return;
  fprintf(client.log.file, "\nLogging player state in time %d :", world.time.clock);
  fprintf(client.log.file, "\n  playmode : %s", player.playmodestr);
  fprintf(client.log.file, "\n  dir : %.2f , locx : %.2f , locy : %.2f",
          world.body.loc.body_dir, world.body.loc.x, world.body.loc.y);
  fprintf(client.log.file, "\n  stamina : %.2f , effort : %.2f", world.body.stamina, world.body.effort);
  fprintf(client.log.file, "\n  speed : %.2f , head_angle : %.2f", speed(world.body.loc), world.body.head_angle);
  fprintf(client.log.file, "\n  ballvx: %.2f , ballvy : %.2f",
          world.ball.vx,world.ball.vy);
  fprintf(client.log.file, "\n  ballx: %.2f , bally : %.2f, dir : %.2f, speed %.2f, crr %.2f",
          world.ball.x, world.ball.y, reldir_2me(world.ball), speed(world.ball), world.ball.correctness);
  fprintf(client.log.file, "\n  offside lines: our %.2f , opp : %.2f",
          world.offside.our.line,world.offside.opp.line);
  fprintf(client.log.file, "\n  estimates\n  ballx: %.2f , bally : %.2f\n",
          estimate_ball_x(1), estimate_ball_y(1));
}

void logcounters(void)
{
  if ((!LOGIT)||(!LOG_ALL_WORLD_ST)) return;
  fprintf(client.log.file, "\nLogging counters time %d (server - player) :", world.time.clock);
  fprintf(client.log.file, "\n  turn : %d - %d", world.counter.turn, world.body.plcount.turn);
  fprintf(client.log.file, "\n  dash : %d - %d", world.counter.dash, world.body.plcount.dash);
  fprintf(client.log.file, "\n  kick : %d - %d", world.counter.kick, world.body.plcount.kick);
  fprintf(client.log.file, "\n");
}

void logobjstates(void)
{
  int i;
  logplayerstate();
  if ((!LOGIT)||(!LOG_ALL_WORLD_ST)) return;
  logcounters();
  fprintf(client.log.file, "\nLogging opponents' state in time %d :", world.time.clock);
  for(i=1;i<=N_PLAYER_PER_TEAM;i++) {
    fprintf(client.log.file, "\n  unum %d - x : %.2f, y : %.2f", i, world.opp[i].x, world.opp[i].y);
  }
  fprintf(client.log.file, "\nLogging teammates' state in time %d :", world.time.clock);
  for(i=1;i<=N_PLAYER_PER_TEAM;i++) {
    fprintf(client.log.file, "\n  unum %d - x : %.2f, y : %.2f", i, world.tmate[i].x, world.tmate[i].y);
  }
  fprintf(client.log.file, "\n");
}
