
#define _skills_dribble_extern_
#include "dribble.h"

c_dribble::c_dribble(void)
{
  maxpower = 100;
  destx = 52;
  desty = 0;
}

void c_dribble::update_bytime(void)
{
  com2send.type = NO_COMMAND;
  calculated = false;
  value = 0;
}

bool c_dribble::possible(void)
{
  if (!can_kickball()) return false;
  return true;
}

double c_dribble::weight(void)
{
  if (!calculated) calculate();
  return value;
}

double c_dribble::cone_val(double dir)
{
  double opp_x,opp_y;
  double angle2opp,angle_dif,opp2ball;
  double dist_val=100,angle_val=100;
  double cone_value;
  double D_CONE_DEPTH=7.0,D_CONE_ANGLE=70;
  int i;
  
  for (i=1;i<=N_PLAYER_PER_TEAM;i++) {
    opp_x = estimate_player_x(world.opp[i],1);
    opp_y = estimate_player_y(world.opp[i],1);
    angle2opp = direction(opp_x-world.ball.x,opp_y-world.ball.y);
    angle_dif = fabs(angle2opp-dir);
    opp2ball  = dist2point(opp_x,opp_y,world.ball);
    if ((opp2ball<D_CONE_DEPTH)&&
        (opp2ball*cos(Deg2Rad(angle_dif))>Pow(opp2ball*sin(Deg2Rad(angle_dif))))) {
      angle_val = angle_val * ((1+angle_dif)/11)/(D_CONE_ANGLE/20);
      dist_val  = dist_val  * opp2ball/D_CONE_DEPTH;
    };
//    if ((opp2ball<D_CONE_DEPTH)&&(angle_dif<D_CONE_ANGLE/2)) {
//      angle_val = angle_val * ((1+angle_dif)/11)/(D_CONE_ANGLE/20);
//      dist_val  = dist_val  * opp2ball/D_CONE_DEPTH;
//    };
  }
  cone_value = 0.5*angle_val + 0.5*dist_val;
  return cone_value;  
}

void c_dribble::find_best_dribble_direction(void)
{
  double cone_value;
  double closeness_value;
  double free_space_value;
  double dribble_value;
  double angle2target,angle_dif; 
  int i,k,c;

  angle2target = direction(destx-world.ball.x,desty-world.ball.y);
  for(c=60;c>=0;c--) {
    for (k=-1;k<=2;k+=2) {
      i = k*c + (int)angle2target;
      //calculating cone value ...
      cone_value = cone_val(i); 
      //calculating closeness value ...
      angle_dif = fabs(i-angle2target);
      closeness_value = 100-angle_dif; // 100 - c;
      if (closeness_value<0) closeness_value = 0;
      //calculating free space value ...
      free_space_value = 5*free_dist_cone(i);
      if (free_space_value>100) free_space_value = 100;
      //calculating final dribble value ...
      dribble_value = 0.70*cone_value + 0.10*closeness_value + 0.20*free_space_value;
      if ((LOGIT)&&(LOG_SKILL)) {
	fprintf(client.log.file,"\n---------------dribble angle-------------- %d",i);
        fprintf(client.log.file,"\n--dribble : cone value          : %.2f",cone_value);
        fprintf(client.log.file,"\n--dribble : closeness value     : %.2f",closeness_value);
        fprintf(client.log.file,"\n--dribble : free space value    : %.2f",free_space_value);
        fprintf(client.log.file,"\n--dribble : final dribble value : %.2f",dribble_value);
      }
      if (dribble_value > value) {
	value = dribble_value;
        best_direction = i;
      }
    }  
  }
}

int c_dribble::too_close_opp(double dir)
{
  double opp2ball, angle2opp, angle_dif;
  double min_dist=FIELD_LENGTH;
  int i, opponent=0;

  for (i=1;i<=N_PLAYER_PER_TEAM;i++) {
    angle2opp = direction(world.ball,world.opp[i]);
    angle_dif = fabs(angle2opp-dir);
    opp2ball = dist2point(world.opp[i],world.ball);
    if ((opp2ball<min(min_dist,DRIBBLE_DANGER_RADIUS))&&(angle_dif<45)) {
      min_dist=opp2ball;
      opponent=i;
    }       
  }
  return opponent;
}

void c_dribble::spiral_dribble(int num)
{
  if (sin(Deg2Rad(reldir_2me(world.opp[num])-reldir_2me(world.ball)))>0)
    spin(com2send,1);
  else 
    spin(com2send,-1);
}

void c_dribble::perform(void)
{
  if (!calculated) calculate();

  if ((LOGIT)&&(LOG_SKILL)) {
    fprintf(client.log.file, "\nLogging Dribble state in time %d :", world.time.clock);
    fprintf(client.log.file, "\n  Skill : Dribbling the opponents in direction %.2f", best_direction);
    fprintf(client.log.file, "\n    Using dribble case : %d\n", dribble_case);
  };

  last_dribble_time = world.time.clock;

  perform_command(com2send);
}

void c_dribble::calculate(void)
{
  double dribble_shoot_factor;
  double dx, dy;
  int dangerous_opp;
  
  find_best_dribble_direction();
  dangerous_opp = too_close_opp(best_direction);
  if (free_dist_cone(best_direction)<5) {
    dribble_case = 1;
    dribble_shoot_factor=1.7;//1.7
    dx = dribble_shoot_factor*free_dist_cone(best_direction)*cos(Deg2Rad(best_direction));
    dy = dribble_shoot_factor*free_dist_cone(best_direction)*sin(Deg2Rad(best_direction));
    if ((LOGIT)&&(LOG_SKILL)) {
      fprintf(client.log.file, "\n-- D/F : free dist cone : %.2f",free_dist_cone(best_direction));
      fprintf(client.log.file, "\n-- D/F : dx,dy,dist     : %.2f , %.2f , %.2f",dx,dy,dist(dx,dy));
    }
    if (!kick2location(com2send, world.ball.x+dx,world.ball.y+dy,0.5))
      kick2direction(com2send,best_direction,50);
    if (collide(com2send)) spin_kick(com2send, dir_from_me(dx,dy));
  } else if (dangerous_opp!=0) { 
    dribble_case = 2;
    if ((LOGIT)&&(LOG_SKILL)) {
      fprintf(client.log.file, "-- D/S : dangerous_opp  : %d",dangerous_opp);
    }
    spiral_dribble(dangerous_opp);
  } else { 
    dribble_case = 3;
    dribble_shoot_factor=0.3;//0.3
    if ((Pow((52-world.body.loc.x)/20))+(Pow(world.body.loc.y/25))<1) dribble_shoot_factor=0.1;//0.1
//    if (world.body.loc.x>36) dribble_shoot_factor=0.15;
    dx = dribble_shoot_factor*free_dist_cone(best_direction)*cos(Deg2Rad(best_direction));
    dy = dribble_shoot_factor*free_dist_cone(best_direction)*sin(Deg2Rad(best_direction));
    if ((LOGIT)&&(LOG_SKILL)) {
      fprintf(client.log.file, "\n-- D/T : dribble_factor : %.2f\n",dribble_shoot_factor);
    }
    if ((world.ball.x+dx>33)&&(world.ball.x<30)) dx = 33-world.ball.x; 
    if (dist(dx,dy)<6) {  
      if (!kick2location(com2send,world.ball.x+dx,world.ball.y+dy,0.5))
        kick2velocity(com2send,best_direction,0.65);
    }
    else kick2velocity(com2send,best_direction,0.65);
    if (collide(com2send)) spin_kick(com2send, dir_from_me(dx,dy));
  }

  calculated = true;
}

