/*
  SKILLS : SOURCE FILE
  
  Skills utilities: 
    kick ball to reach a location
    kick ball with a power in a direction
    kick ball to get to a velocity
    do effective turn

  Ehsan Foroughi 

  ParsAI Team
  Sharif University of Tech.

 */

#define _skills_utils_extern_
#include "utils.h"

double effective_turn(double dir)
{
  double moment;

  moment = dir*(1.0+INERTIA_MOMENT*speed(world.body.loc));
  if (moment>180) moment = 180;
  if (moment<-180) moment = -180;

  return moment;  
}

double needed_vx4kick(double x, double y, double fv)
{
  double dx, dy, vx1, t;
  
  dx = x - world.ball.x;
  dy = y - world.ball.y;
  if (equal_float(dx,0)) return ((double)0);
  
  vx1 = fv * dx / dist(dx,dy);
  t = BALL_DECAY + dx * (1-BALL_DECAY) / vx1;
  return (vx1*t);
}

double needed_vy4kick(double x, double y, double fv)
{
  double dx, dy, vy1, t;
  
  dx = x - world.ball.x;
  dy = y - world.ball.y;
  if (equal_float(dy,0)) return ((double)0);
  
  vy1 = fv * dy / dist(dx,dy);
  t = BALL_DECAY + dy * (1-BALL_DECAY) / vy1;
  return (vy1*t);
}

bool kick2location(Command &com, double x, double y, double fv)
{
  double dvx, dvy, power_rate, power, dir;  

  if (equal_float(dist(x-world.ball.x,y-world.ball.y),0)) return true;
  if (!can_kickball()) return false;
  if (fv<LEAST_FIN_BALL_VEL) fv = LEAST_FIN_BALL_VEL;
  
  dvx = needed_vx4kick(x,y,fv)-world.ball.vx;
  dvy = needed_vy4kick(x,y,fv)-world.ball.vy;
  
  power_rate = KICK_POWER_RATE * (1 - 0.25 * fabs(reldir_2me(world.ball))/180
    - 0.25 * (dist_2me(world.ball)-PLAYER_SIZE-BALL_SIZE)/KICKABLE_MARGIN);
  power = dist(dvx,dvy)/power_rate;
  dir = normalize_angle(Rad2Deg(atan2(dvy,dvx))-world.body.loc.body_dir);

  if ((LOGIT)&&(LOG_SKILL)) {
    fprintf(client.log.file,"Kick2Location : x=%.2f,y=%.2f,fv=%.2f \n", x,y,fv);
  };
  
  if (power>100) 
    return false;

  if ((LOGIT)&&(LOG_SKILL)) {
    fprintf(client.log.file,"Kick2Location : kicked with power %.2f\n", power);
  };

  //if (power>100) power = 100;
  com.type = KICK;
  com.par1 = power;
  com.par2 = dir;
//  kick(power,dir);

  return true;
}


bool one_cycle_kick2location(Command &com,double x,double y)                                                                 
{                                                                                                                            
  double dvx, dvy, power_rate, kickpower, fdir;                                                                              
                                                                                                                             
  if (!can_kickball()) return false;                                                                                         
                                                                                                                             
  dvx = x-world.ball.x-world.ball.vx;                                                                                        
  dvy = y-world.ball.y-world.ball.vy;                                                                                        
                                                                                                                             
  power_rate = KICK_POWER_RATE * (1 - 0.25 * fabs(reldir_2me(world.ball))/180                                                
    - 0.25 * (dist_2me(world.ball)-PLAYER_SIZE-BALL_SIZE)/KICKABLE_MARGIN);                                                  
  kickpower = dist(dvx,dvy)/power_rate;                                                                                      

  if ((LOGIT)&&(LOG_SKILL)) {
    fprintf(client.log.file,"OneCycleKick2Location : x=%.2f,y=%.2f\n", x,y);
  };
                                                                                                                       
  if (kickpower>100) return false;                                                                                           

  if ((LOGIT)&&(LOG_SKILL)) {
    fprintf(client.log.file,"OneCycleKick2Location : kicked with power %.2f\n", kickpower);
  };

  fdir = normalize_angle(Rad2Deg(atan2(dvy,dvx))-world.body.loc.body_dir);                                                   
                                                                                                                             
  com.type = KICK;                                                                                                           
  com.par1 = kickpower;                                                                                                      
  com.par2 = fdir;                                                                                                           
  return true;                                                                                                               
}                     


bool kick2direction(Command &com, double dir, double kickpower)
{
  double vx, vy, power_rate, power, cd, rd, brd, kdir;  

  if (!can_kickball()) return false;
  if ((LOGIT)&&(LOG_SKILL)) {
    fprintf(client.log.file,"Kick2direction : dir=%.2f power=%.2f \n",dir,kickpower);
  };

  
  if (kickpower>90) return (kick2velocity(com,dir,5.0));
  vx = world.ball.vx;
  vy = world.ball.vy;

  if (vx < 0.1) vx = 0;
  if (vy < 0.1) vy = 0;

  power_rate = KICK_POWER_RATE * (1 - 0.25 * fabs(reldir_2me(world.ball))/180
    - 0.25 * (dist_2me(world.ball)-PLAYER_SIZE-BALL_SIZE)/KICKABLE_MARGIN);
  power = kickpower*power_rate;

  kdir = dir;
  brd = Rad2Deg(atan2(vy+power*sin(Deg2Rad(kdir)),vx+power*cos(Deg2Rad(kdir))));
  for (cd=-180;cd<180;cd+=0.2) {
    rd = Rad2Deg(atan2(vy+power*sin(Deg2Rad(cd)),vx+power*cos(Deg2Rad(cd))));
    if (fabs(normalize_angle(rd-dir))<fabs(normalize_angle(brd-dir))) {
      kdir = cd;
      brd = rd;
    }
  }

  if ((LOGIT)&&(LOG_SKILL)) {
    fprintf(client.log.file,
    "Kick2Dir : needed_dir %.2f , shoot_dir %.2f , bvx %.2f, bvy %.2f", dir, kdir, vx, vy);
  };


  com.type = KICK;
  com.par1 = kickpower;
  com.par2 = normalize_angle(kdir-world.body.loc.body_dir);
//  kick(kickpower,normalize_angle(kdir-world.body.loc.body_dir));
  return true;
}

bool kick2velocity(Command &com, double dir, double sp)
{
  double dvx, dvy, power_rate, kickpower, fdir;
  if (!can_kickball()) return false;

  dvx = cos(Deg2Rad(dir))*sp-world.ball.vx;
  dvy = sin(Deg2Rad(dir))*sp-world.ball.vy;

  power_rate = KICK_POWER_RATE * (1 - 0.25 * fabs(reldir_2me(world.ball))/180
    - 0.25 * (dist_2me(world.ball)-PLAYER_SIZE-BALL_SIZE)/KICKABLE_MARGIN);
  kickpower = dist(dvx,dvy)/power_rate;
  if ((LOGIT)&&(LOG_SKILL)) {
    fprintf(client.log.file,"Kick2Velocity : dir=%.2f,sp=%.2f,power=%.2f \n", dir,sp,kickpower);
  };

#ifdef DEBUG_ON
  sprintf(debugbuf,"Kick2Velocity : corrected - dir=%.2f,sp=%.2f\n,kickpower=%.2f\n", dir,sp,kickpower);
  logmsg(debugbuf);
#endif

  fdir = normalize_angle(Rad2Deg(atan2(dvy,dvx))-world.body.loc.body_dir);

  if (kickpower>100) 
  {
    if (sp<0.05) // is that a stop ball?! 
    {  
       // can't stop ball? just kick against with maxpower;
        kickpower=100;
                                                                                                                             
#ifdef DEBUG_ON
  sprintf(debugbuf,"Kick2velocity :  Kickagainst\n");
  logmsg(debugbuf);
#endif

    }
    else // kick with less power but in the same dir
    {
      double a,b,c,k;
      a=1;
      b=-2*(world.ball.vx*cos(Deg2Rad(dir))+world.ball.vy*sin(Deg2Rad(dir)));      c=Pow(speed(world.ball))-Pow(100*power_rate);
      if (solve(a,b,c,k)) // is there a way to kick with less power in same dir?
      { 
        sp=k;
        dvx = cos(Deg2Rad(dir))*sp-world.ball.vx;
        dvy = sin(Deg2Rad(dir))*sp-world.ball.vy;
        power_rate = KICK_POWER_RATE * (1 - 0.25 * fabs(reldir_2me(world.ball))/180
                     - 0.25 * (dist_2me(world.ball)-PLAYER_SIZE-BALL_SIZE)/KICKABLE_MARGIN);
        kickpower = dist(dvx,dvy)/power_rate;

#ifdef DEBUG_ON
  sprintf(debugbuf,"Kick2Velocity : corrected - dir=%.2f,sp=%.2f\n,kickpower=%.2f\n", dir,sp,kickpower);
  logmsg(debugbuf);
#endif

        fdir = normalize_angle(Rad2Deg(atan2(dvy,dvx))-world.body.loc.body_dir);
      }
      else { // no?! just stop ball
        kick2velocity(com,0,0);
#ifdef DEBUG_ON
  sprintf(debugbuf,"Kick2Velocity : cannot correct! stop the ball\n");
  logmsg(debugbuf);
#endif

      }
    }
  }
  

  com.type = KICK;
  com.par1 = kickpower;
  com.par2 = fdir;

//  kick(kickpower,fdir);
  return true;
}

void spin(Command &com, int rotate) 
{
  double ang,bdist,posx,posy;

  bdist = PLAYER_SIZE+3.0*BALL_SIZE;
  ang   = world.body.loc.body_dir+reldir_2me(world.ball)-rotate*30;
  posx  = world.body.loc.x+bdist*cos(Deg2Rad(ang))+world.body.loc.vx;
  posy  = world.body.loc.y+bdist*sin(Deg2Rad(ang))+world.body.loc.vy;

  if ((LOGIT)&&(LOG_SKILL)) {

    fprintf(client.log.file,"  Spin %s to x=%.2f, y=%.2f\n",(rotate==1)?"CCW":"CW",posx,posy);
  };

  if (fabs(fabs(world.ball.y)-FIELD_WIDTH/2)<0.5) kick2direction(com, direction(0,0),25);
  else if (!kick2location(com, posx,posy,0.4))
  if (!kick2location(com, posx,posy,0.3)) kick2velocity(com,0,0);
}
/*
void spin(Command &com, int rotate) 
{
  double ang,bdist,posx,posy;

  bdist = PLAYER_SIZE+3.0*BALL_SIZE;
  ang   = world.body.loc.body_dir+reldir_2me(world.ball)-rotate*30;
  posx  = world.body.loc.x+bdist*cos(Deg2Rad(ang))+world.body.loc.vx;
  posy  = world.body.loc.y+bdist*sin(Deg2Rad(ang))+world.body.loc.vy;

  if ((LOGIT)&&(LOG_SKILL)) {

    fprintf(client.log.file,"  Spin %s to x=%.2f, y=%.2f\n",(rotate==1)?"CCW":"CW",posx,posy);
  };

  if (fabs(fabs(world.ball.y)-FIELD_WIDTH/2)<0.5) kick2direction(com, direction(0,0),25);
  else if (!one_cycle_kick2location(com,posx,posy))
  //else if (!kick2location(com, posx,posy,0.4))
  //if (!kick2location(com, posx,posy,0.3)) 
  kick2velocity(com,direction(posx-world.ball.x,posy-world.ball.y),3.0);
}
*/
// it will kick 2 a dir relative to my body
void spin_2dir(Command &com, double dir)
{
  double ang,bdist,posx,posy;

  bdist = PLAYER_SIZE+3.0*BALL_SIZE;
  ang   = world.body.loc.body_dir+dir;
  posx  = world.body.loc.x+bdist*cos(Deg2Rad(ang))+world.body.loc.vx;
  posy  = world.body.loc.y+bdist*sin(Deg2Rad(ang))+world.body.loc.vy;
  if ((LOGIT)&&(LOG_SKILL)) {

    fprintf(client.log.file,"  Spin_2dir dir=%.2f x=%.2f, y=%.2f\n",dir,posx,posy);
  };

  if (fabs(fabs(world.ball.y)-FIELD_WIDTH/2)<0.5) kick2direction(com, direction(0,0),25);
//  if (!one_cycle_kick2location(com,posx,posy))
  if (!kick2location(com,posx,posy,0.3))
    kick2velocity(com,direction(posx-world.ball.x,posy-world.ball.y),3.0);
  
}

// it will kick 2 a dir relative to my body
void spin_to(Command &com, double dir,double dist)
{
  double ang,bdist,posx,posy;

  bdist = dist;
  ang   = world.body.loc.body_dir+dir;
  posx  = world.body.loc.x+bdist*cos(Deg2Rad(ang))+world.body.loc.vx;
  posy  = world.body.loc.y+bdist*sin(Deg2Rad(ang))+world.body.loc.vy;

  if ((LOGIT)&&(LOG_SKILL)) {

    fprintf(client.log.file,"  Spin_2to to dir=%.2f , dist = %.2f x=%.2f, y=%.2f\n",dir,dist,posx,posy);
  };

  if (fabs(fabs(world.ball.y)-FIELD_WIDTH/2)<0.5) kick2direction(com, direction(0,0),25);
//  if (!one_cycle_kick2location(com,posx,posy))
  if (!kick2location(com,posx,posy,0.3))
  {
    kick2velocity(com,0,0);
  }
  
}


void spin_kick(Command &com, double dir) 
{
  double d;
  int num;

  d = normalize_angle(dir_from_me(world.ball) - dir);
  num = nearmost_player(world.body.loc, OPPONENTS);

  if (fabs(d)>(45*(dist_2me(world.opp[num])-0.7))+90)
    d = -normalize_angle(dir_from_me(world.opp[num]) - dir);
  if ((LOGIT)&&(LOG_SKILL)) {

    fprintf(client.log.file,"  Spin_kick to dir=%.2f \n",dir);
  };

  if (d>0) spin(com,1);
  else spin(com,-1);
}

// By Pooya - point p3 distance from line p1-p2                                                     
double point_2line_dist(double x1,double y1,double x2,double y2,double x3,double y3)                
{                                                                                                   
  double a,b,c;                                                                                     
  a=y2-y1;                                                                                          
  b=x1-x2;                                                                                          
  c=-1.0*y2*x1+x2*y1;                                                                               
  if (sqrt((Pow(a)+Pow(b)))==0.0)                                                                   
    return 1000.0;                                                                                  
  return fabs(a*x3+b*y3+c)/sqrt((Pow(a)+Pow(b)));                                                   
}                                                                                                   
                                                                                                    
                                                                                                    
// By Pooya                                                                                         
// x=0    --> y=0                                                                                   
// x=a    --> y=b                                                                                   
// x->inf --> y=0                                                                                   
double fuzzy(double x,double a,double b)                                                            
{                                                                                                   
  double v;                                                                                         
  v=(3*b*Pow(x))/(2*pow(x,3)/a+Pow(a));                                                             
  return v;                                                                                         
}                    

// By Pooya - solve a*x^2+b*x+c=0 and return the x>0 answers
bool solve(double a,double b,double c,double &answer)
{
  double delta;
  delta=b*b-4*a*c;
  if (delta<0)
  {
    return false;
  }
  else
  {
    answer=(-b+sqrt(delta))/(2*a);
  }
  if (answer<0) return false;
  return true;
}

bool is_in_field(double x, double y)
{
  if (x>FIELD_LENGTH/2) return false;
  if (x<-FIELD_LENGTH/2) return false;
  if (y>FIELD_WIDTH/2) return false;
  if (y<-FIELD_WIDTH/2) return false;
  return true;
}

bool is_in_field(Objpos &p)
{
  return (is_in_field(p.x, p.y));
}

