/*
  Low level utilities: 
    random functions
    relative cordination functions
    angle calculation functions

  Ehsan Foroughi 
  ParsAI Team
  Sharif University of Tech.

 */

#define _utils_extern_
#include "utils.h"

void swap(double &x,double &y)
{
  double t;
  t=x;x=y;y=t;
}

void swap(int &x,int &y)
{
  int t;
  t=x;x=y;y=t;
}

int genrand(int n)
{
  return((int)(1.0*n*rand()/(RAND_MAX+1.0)));
}

bool equal_float(double a,double b)
{
  return(fabs(a-b)<FLOAT_ERROR);
}

bool equal_dist(double a, double b)
{
  return(fabs(a-b)<LOC_ERROR);
}

double normalize_angle(double angle)
{
  double retval;
  /* first make sure |retval| < 360 */
  retval = angle - (int)(angle/360.)*360;
  if (retval > 180.0) {
    retval -= 360.0;
  } else if (retval < -180.0) {
    retval += 360.0;
  }
  return retval;
}

double diff_angle(double ang1, double ang2)
{
  return (fabs(normalize_angle(ang1-ang2)));
}

double dist(double dx, double dy)
{
  return(sqrt(Pow(dx)+Pow(dy)));
}

double dist(Objpos &pos, double x, double y)
{
  return(dist(pos.x-x,pos.y-y));
}

double dist(Pos &pos, Objpos &obj)
{
  return(dist(pos.x-obj.x,pos.y-obj.y));
}

double dist(Objpos &pos, Objpos &obj)
{
  return(dist(pos.x-obj.x,pos.y-obj.y));
}

double dist2point(double px,double py,Objpos &obj)
{
  return dist(px-obj.x,py-obj.y);
}

double dist2point(Objpos &objs,Objpos &objd)
{
  return dist(objs.x-objd.x,objs.y-objd.y);
}

double dist_2me(double x, double y)
{
  return(sqrt(Pow(x-world.body.loc.x)+Pow(y-world.body.loc.y)));
}

double dist_2me(Objpos &obj)
{
  return(sqrt(Pow(obj.x-world.body.loc.x)+Pow(obj.y-world.body.loc.y)));
}

double direction(double dx, double dy)
{
  return(normalize_angle(Rad2Deg(atan2(dy,dx))));
}

double direction(Objpos &objs, Objpos &objd)
{
  return(direction(objd.x-objs.x, objd.y-objs.y));
}

double dir_from_me(double x, double y)
{
  return(direction(x-world.body.loc.x, y-world.body.loc.y));
}

double dir_from_me(Objpos &obj)
{
  return(dir_from_me(obj.x, obj.y));
}

double reldir_2me(double x, double y)
{
  double t = Rad2Deg(atan2(y-world.body.loc.y,x-world.body.loc.x));
  return(normalize_angle(t-world.body.loc.body_dir));
}

double reldir_2me(Objpos &obj)
{
  double t = Rad2Deg(atan2(obj.y-world.body.loc.y,obj.x-world.body.loc.x));
  return(normalize_angle(t-world.body.loc.body_dir));
}

double relviewdir_2me(double x, double y)
{
  double t = Rad2Deg(atan2(y-world.body.loc.y,x-world.body.loc.x));
  return(normalize_angle(t-(world.body.loc.body_dir+world.body.head_angle)));
}

double relviewdir_2me(Objpos &obj)
{
  double t = Rad2Deg(atan2(obj.y-world.body.loc.y,obj.x-world.body.loc.x));
  return(normalize_angle(t-(world.body.loc.body_dir+world.body.head_angle)));
}

double speed(Objpos &obj)
{
  return(dist(obj.vx,obj.vy));
}

void com_turn(Command &com, double moment)
{
  com.type = TURN;
  com.par1 = moment;
}

void com_dash(Command &com, double power)
{
  com.type = DASH;
  com.par1 = power;
}

void com_kick(Command &com, double power, double dir)
{
  com.type = KICK;
  com.par1 = power;
  com.par2 = dir;
}

bool is_play_on(void)
{
  switch (player.playmode) {
    case PM_goal_l:
    case PM_goal_r:
    case PM_offside_l:
    case PM_offside_r:
    case PM_before_kick_off:
      return false;
    default: return true;
  };
}

Side play_mode_side(PlayMode pm)
{
  Side ret = S_UNKNOWN;

  if (strncmp("offside", pm_entry[pm].name, strlen("offside"))) {
    if (pm_entry[pm].name[strlen(pm_entry[pm].name)-1]=='r') ret = S_RIGHT;
    else if (pm_entry[pm].name[strlen(pm_entry[pm].name)-1]=='l') ret = S_LEFT;
  } else {
    if (pm_entry[pm].name[strlen(pm_entry[pm].name)-1]=='r') ret = S_LEFT;
    else if (pm_entry[pm].name[strlen(pm_entry[pm].name)-1]=='l') ret = S_RIGHT;
  }

  return ret;
}

bool str_play_mode(PlayMode pm, char *str)
{
  return(strstr(pm_entry[pm].name,str)!=NULL);
}

bool kick_play_mode(void)
{
  return(str_play_mode(player.playmode,"kick")||str_play_mode(player.playmode,"offside"));
}
