#ifndef _WORLD_OBJECTS_H
#define _WORLD_OBJECTS_H

#ifndef _world_objects_extern_
#define _world_objects_extern_ extern
#endif

#include "../headers.h"

// - Macros
#define moving_obj_negsys(x) ((x>0)?world.tmate[x]:((x<0)?world.opp[-x]:world.ball))
#define moving_obj_possys(x) ((x>N_PLAYER_PER_TEAM)?world.opp[x-N_PLAYER_PER_TEAM]:((x>0)?world.tmate[x]:world.ball))

// - Types of objects

typedef struct _Player {
  Side side;
  int unum;
  char playmodestr[32];
  PlayMode playmode;
  int idtype;
} Player;

typedef struct _Client {
  Socket sock;
  Socket monitor_sock;
  char *server;
  int port;
  char teamname[32];
  Role role;
  struct {
    FILE *file;
    char filename[64];
  } log;
} Client;

typedef struct _World {
  Time time, lasttime;
  bool time_running;
  struct {
    Time   last_catch;
    bool   ball_near_sensor;
    bool   ball_inhand;
    Objpos loc;
    double stamina;
    double effort;
    PVector speed;
    double head_angle;
    Counter plcount;
    bool   heardref;
  } body;
  Counter counter;
  ViewState view;
  Objpos opp[N_PLAYER_PER_TEAM+1],
         tmate[N_PLAYER_PER_TEAM+1],
         ball, ball_last_cycle;
  struct {
    offdata our, opp;
    bool inoff;
  } offside;
  struct {
    bool incontrol;
    bool getball;
    int last_kickable_cycle;
  } control_ball;
} World;

typedef struct _Sensor {
  Command sent_commands[SENT_COMMANDS_MEM];
  SenseBodyInfo body;
  SeeInfo see;
  HearInfo hear;
  bool lost_message;
  struct {
    bool sth2say;
    int passdest;
  } comm;  
} Sensor;

typedef struct _FullstateInfo {
  Objpos opp[N_PLAYER_PER_TEAM+1],
         tmate[N_PLAYER_PER_TEAM+1],
         ball;
  Score score;
  char playmodestr[32];
  PlayMode playmode;
  ViewState view;
  Time time;
} FullstateInfo;

// - Objects
_world_objects_extern_ Player player;
_world_objects_extern_ Client client;
_world_objects_extern_ World world;
_world_objects_extern_ Sensor sensor;
_world_objects_extern_ FullstateInfo fullstate;

_world_objects_extern_ char debugbuf[512];

#endif
