#define _world_sensor_extern_
#include "sensor.h"

void estimate_object_pos(PosState me, double object_dist, double object_dir, PosState *obj_info)
{
  (obj_info->pos).x = me.pos.x + object_dist * cos(Deg2Rad(me.dir+object_dir));
  (obj_info->pos).y = me.pos.y + object_dist * sin(Deg2Rad(me.dir+object_dir));
}

void update_commands_from_time(int tclock)
{
/*  int i;  
  if (world.time.clock<tclock) return;
  i = timing_index(world.time);
  while (sent_commands[i].clock>tclock) {
    if (i==0) i = SENT_COMMANDS_MEM;
    i--;
  }
  while (i<=timing_index(world.time)) {
  };*/
}

// - Checks if the unseen object should have seen ( Ghost )
void check_ghost(Objpos &obj)
{
  if (fabs(relviewdir_2me(obj))>view_angle()-GHOST_CARE_ANGLE) return;
  if (dist_2me(obj)>VISIBLE_DISTANCE-0.5) return; 
  // This is ghost
  obj.correctness = GHOST_CORRECTNESS;
}

// - Replaces the unum and the team from memory if not have been seen (e.g. too far)
void find_unum(PlayerInfo &pl, Pos &pos)
{
  int i, uo=0, ut=0;
  if (pl.team!=WT_Opponent) {
    ut=1;
    for (i=1;i<=N_PLAYER_PER_TEAM;i++)
      if (dist(pos,world.tmate[i])<dist(pos,world.tmate[ut])) ut=i;
    if (dist(pos,world.tmate[ut])>MAX_UNUM_FIND_DIST) ut=0;
  };
  if (pl.team!=WT_Our) {
    uo=1;
    for (i=1;i<=N_PLAYER_PER_TEAM;i++)
      if (dist(pos,world.opp[i])<dist(pos,world.opp[uo])) uo=i;
    if (dist(pos,world.opp[uo])>MAX_UNUM_FIND_DIST) uo=0;
  };
  if (uo==0) {pl.unum=ut;pl.team=WT_Our;}
  else if (ut==0) {pl.unum=uo;pl.team=WT_Opponent;}
  else if (dist(pos,world.tmate[ut])<dist(pos,world.opp[uo])) {pl.unum=ut;pl.team=WT_Our;}
  else {pl.unum=uo;pl.team=WT_Opponent;};
}

void update_world_bysee(SeeInfo *see)
{
  PosState gst, mypos;
  bool tmate_ghost[N_PLAYER_PER_TEAM+1],
       opp_ghost[N_PLAYER_PER_TEAM+1],
       ball_ghost;
  int i, j;

  if (see->time>=sensor.see.time) sensor.see = *see;
  else return;

  for (i=0;i<=N_PLAYER_PER_TEAM;i++) {
    tmate_ghost[i] = true;
    opp_ghost[i] = true;
  }
  ball_ghost = true;

  // - Find my position and direction in field
  if (GOOD_POSITIONING) {
    if (!estimate_current_pos5(see,player.side,&mypos)) return;
  } else {
    if (estimate_current_pos3(see,player.side,&gst)) mypos = gst;
    else if (estimate_current_pos2(see,player.side,&gst)) mypos = gst;
    else if (estimate_current_pos(see,player.side,&gst)) mypos = gst;
    else return;
  };

#ifdef DEBUG_ON
    sprintf(debugbuf,"! direction : %.2f", mypos.dir-world.body.head_angle);
    logmsg(debugbuf);
#endif

  // - Updating myself world model
  world.body.loc.correctness = SEE_CORRECTNESS;
  world.body.loc.body_dir = mypos.dir-world.body.head_angle;
  world.body.loc.x = mypos.pos.x;
  world.body.loc.y = mypos.pos.y;
  
  world.tmate[player.unum].x = mypos.pos.x;
  world.tmate[player.unum].y = mypos.pos.y;
  world.tmate[player.unum].body_dir = mypos.dir-world.body.head_angle;
  world.tmate[player.unum].correctness = SEE_CORRECTNESS;
  world.tmate[player.unum].update_time = world.time;
  tmate_ghost[player.unum] = false;

  for(i=0;i<see->n_obj;i++)
    if ((see->obj[i].type==OT_Player)&&(!equal_float(see->obj[i].dist,DIST_ERR))) {
      if (see->obj[i].id.player.team==WT_Opponent) {
        estimate_object_pos(mypos,see->obj[i].dist,see->obj[i].dir,&gst);
        j=see->obj[i].id.player.unum;
        if ((j<=0)||(j>N_PLAYER_PER_TEAM)) find_unum(see->obj[i].id.player,gst.pos);
        if ((j>0)&&(j<=N_PLAYER_PER_TEAM)) {
          world.opp[j].x=gst.pos.x;
          world.opp[j].y=gst.pos.y;
          world.opp[j].face_dir=see->obj[i].face_dir;
          world.opp[j].body_dir=see->obj[i].body_dir;
          if ((!equal_float(see->obj[i].dir_chng,CHNG_ERR))&&(!equal_float(see->obj[i].dist_chng,CHNG_ERR))) {
            world.opp[j].vx=get_vx(gst.pos.x,gst.pos.y,see->obj[i].dist_chng,see->obj[i].dir_chng);
            world.opp[j].vy=get_vy(gst.pos.x,gst.pos.y,see->obj[i].dist_chng,see->obj[i].dir_chng);
          };
          world.opp[j].correctness=SEE_CORRECTNESS-CORRECTNESS_DIST_W*dist_2me(world.opp[j]);
          world.opp[j].update_time=world.time;
          opp_ghost[j] = false;
	};
      } else if (see->obj[i].id.player.team==WT_Our) {
        estimate_object_pos(mypos,see->obj[i].dist,see->obj[i].dir,&gst);
        j=see->obj[i].id.player.unum;
        if ((j>0)&&(j<=N_PLAYER_PER_TEAM)) {
          world.tmate[j].x=gst.pos.x;
          world.tmate[j].y=gst.pos.y;
          world.tmate[j].face_dir=see->obj[i].face_dir;
          world.tmate[j].body_dir=see->obj[i].body_dir;
          if ((!equal_float(see->obj[i].dir_chng,CHNG_ERR))&&(!equal_float(see->obj[i].dist_chng,CHNG_ERR))) {
            world.tmate[j].vx=get_vx(gst.pos.x,gst.pos.y,see->obj[i].dist_chng,see->obj[i].dir_chng);
            world.tmate[j].vy=get_vy(gst.pos.x,gst.pos.y,see->obj[i].dist_chng,see->obj[i].dir_chng);
          };
          world.tmate[j].correctness=SEE_CORRECTNESS-CORRECTNESS_DIST_W*dist_2me(world.tmate[j]);
          world.tmate[j].update_time=world.time;
          tmate_ghost[j] = false;
	};
      };
    } else if (see->obj[i].type==OT_Ball) {
      estimate_object_pos(mypos,see->obj[i].dist,see->obj[i].dir,&gst);
      world.ball.x=gst.pos.x;
      world.ball.y=gst.pos.y;
      if ((!equal_float(see->obj[i].dir_chng,CHNG_ERR))&&(!equal_float(see->obj[i].dist_chng,CHNG_ERR))) {
        world.ball.vx=get_vx(gst.pos.x,gst.pos.y,see->obj[i].dist_chng,see->obj[i].dir_chng);
        world.ball.vy=get_vy(gst.pos.x,gst.pos.y,see->obj[i].dist_chng,see->obj[i].dir_chng);
      }; 

      if (see->obj[i].inViewConeP) world.body.ball_near_sensor=false;
      else world.body.ball_near_sensor=true;

      world.ball.correctness=SEE_CORRECTNESS-CORRECTNESS_DIST_W*dist_2me(world.ball);
      world.ball.update_time=world.time;
      ball_ghost = false;
    };
  
  // -- Ghost Remover
  if (ball_ghost) check_ghost(world.ball);
  for (i=1;i<=N_PLAYER_PER_TEAM;i++) {
    if (tmate_ghost[i]) check_ghost(world.tmate[i]);
    if (opp_ghost[i]) check_ghost(world.opp[i]);
  }

  update_commands_from_time(see->time);
}

void get_worldmodel_byhear(char *msg)
{
  char *s;
  int sayer_unum, sayer_stamina, say_time;
  int un, cr;
  double x, y, vx, vy;

  s = msg;

  if (client.role==COACH) return;

  sscanf(s, "%d %d %d", &sayer_unum, &say_time, &sayer_stamina);
  world.tmate[sayer_unum].stamina = sayer_stamina;
  
  nextseptoken(s,'*');  

  sscanf(s, "%d %lf %lf %lf %lf", &cr, &x, &y, &vx, &vy);
  if (world.ball.correctness < cr - HEAR_CRR_DECAY) {
    world.ball.correctness = cr - HEAR_CRR_DECAY;
    world.ball.x = x;
    world.ball.y = y;
    world.ball.vx = vx;
    world.ball.vy = vy;
  };

  nextseptoken(s,'*');  

  while (*s!=0) {
    sscanf(s, "%d %d %lf %lf", &un, &cr, &x, &y);
    nextseptoken(s,'*');  

    if (un==0) {
      if (world.ball.correctness < cr - HEAR_CRR_DECAY) {
        world.ball.correctness = cr - HEAR_CRR_DECAY;
	world.ball.x = x;
	world.ball.y = y;
      };
    } else if (un>0) {
      switch (un) {
       case 12:
        if (cr==player.unum) world.control_ball.getball=true;
	else world.control_ball.getball=false;
        break;
       default:
        if (world.tmate[un].correctness < cr - HEAR_CRR_DECAY) {
          world.tmate[un].correctness = cr - HEAR_CRR_DECAY;
   	  world.tmate[un].x = x;
          world.tmate[un].y = y;
        };
      };
    } else if (un<0) {
      if (world.opp[-un].correctness < cr - HEAR_CRR_DECAY) {
        world.opp[-un].correctness = cr - HEAR_CRR_DECAY;
	world.opp[-un].x = x;
	world.opp[-un].y = y;
      };
    }
  };
}

void get_fromcoach_byhear(char *msg)
{
  char *s;
  int un, n;

  s = msg;

  if (client.role==COACH) return;

  nextseptoken(s,'\"');
  
  for(un=0;un<=22;un++) {
    n = sscanf(s, "%lf %lf", &(moving_obj_possys(un).x), &(moving_obj_possys(un).y));
    assert(n==2);
    skiptokens(s,2);
  };
}

void update_world_byhear(HearInfo *hear)
{
  PlayMode refpm;
  char c;
  int i;

  if (hear->sender==ST_Referee) {
    if((refpm = get_playmode(hear->message))!=PM_Error) {
      if (((refpm==PM_goalie_catch_ball_r)&&(player.side==S_RIGHT))
        ||((refpm==PM_goalie_catch_ball_l)&&(player.side==S_LEFT))) {
	world.body.ball_inhand = true;
	world.ball.x = world.body.loc.x;
	world.ball.y = world.body.loc.y;
        world.ball.vx = 0.0;
        world.ball.vy = 0.0;
        world.ball.update_time = world.time;
      }
//      if (refpm==PM_play_on) world.body.ball_inhand = false;
/*      if (refpm==PM_half_time) {
        if (player.side == S_LEFT) player.side=S_RIGHT;
	else player.side = S_RIGHT;
      };*/
      player.playmode = refpm;
    } else {
      if(!strncmp(hear->message, "goal_", 5)) {
        sscanf(hear->message,"goal_%c_%d",&c,&i);
        if (c=='l') player.playmode = PM_goal_l;
        else player.playmode = PM_goal_r;
	world.ball.x=0;
	world.ball.y=0;
	world.ball.vx=0;
	world.ball.vy=0;
      };
    };
    strcpy(player.playmodestr,pm_entry[player.playmode].name);
    world.time_running = is_play_on();
    world.body.heardref = true;
  } else if (hear->sender==ST_Player) {
    if (!strncmp(hear->message+1, TEAM_SIGNATURE, strlen(TEAM_SIGNATURE)))
      if (!strncmp(hear->message+strlen(TEAM_SIGNATURE)+1, client.teamname, strlen(client.teamname))) {
        if (hear->time>=sensor.hear.time) sensor.hear = *hear;
        get_worldmodel_byhear(hear->message+strlen(TEAM_SIGNATURE)+strlen(client.teamname)+1);
      };
  } else if (hear->sender==((player.side==S_LEFT)?ST_OLCoach_l:ST_OLCoach_r)) {
    if (hear->time>=sensor.hear.time) sensor.hear = *hear;
    get_fromcoach_byhear(hear->message+strlen(TEAM_SIGNATURE)+strlen(client.teamname)+1);
  };
}

void update_world_bybody(SenseBodyInfo *body)
{
  if (body->time>=sensor.body.time) sensor.body = *body;
  else return;

  if ((world.time.clock < body->time)&&(world.time_running==false)&&(!world.body.heardref)) {
    world.time_running = true;
    logmsg("World time begun unexpectedly!! Corrected...");
    report_error("World time begun unexpectedly!? Message lost?");
  } else if ((world.time.clock==body->time)&&(world.time_running==true)&&(!world.body.heardref)) {
    world.time_running = false;
    logmsg("World time stopped unexpectedly!! Corrected...");
    report_error("World time stopped unexpectedly!? Message lost?");
  };
  world.body.heardref = false;

  if (world.time.clock==body->time) world.time.stopped++;
  else if (world.time.clock < body->time){
    world.time.clock = body->time;
  }

  world.body.stamina = body->stamina;
  world.body.effort = body->effort;
  world.counter.kick = body->countKick;
  world.counter.dash = body->countDash;
  world.counter.turn = body->countTurn;
  world.counter.say = body->countSay;
  world.counter.turnneck = body->countTurnNeck;
  world.counter.gcatch = body->countCatch;
  world.counter.move = body->countMove;
  world.counter.change_view = body->countChangeView;
  world.view.quality = body->viewQuality;
  world.view.width = body->viewWidth;
}

void update_world_bysensors(SensorInfo &sense)
{
  if (sense.type==SIT_See)
    update_world_bysee(&(sense.info.see));
  else if (sense.type==SIT_Hear)
    update_world_byhear(&(sense.info.hear));
  else if (sense.type==SIT_SenseBody)
    update_world_bybody(&(sense.info.senseBody));
}
