#define _world_updaters_extern_
#include "updaters.h"

void update_world_kick(double power, double dir)
{
  double power_rate;

  power_rate = KICK_POWER_RATE * (1 - 0.25 * fabs(reldir_2me(world.ball))/180
    - 0.25 * (dist_2me(world.ball)-PLAYER_SIZE-BALL_SIZE)/KICKABLE_MARGIN);
  world.ball.vx += power*power_rate*cos(Deg2Rad(dir+world.body.loc.body_dir));
  world.ball.vy += power*power_rate*sin(Deg2Rad(dir+world.body.loc.body_dir));
}

void update_world_dash(double power)
{
  world.body.loc.vx += power*DASH_POWER_RATE*cos(Deg2Rad(world.body.loc.body_dir));
  world.body.loc.vy += power*DASH_POWER_RATE*sin(Deg2Rad(world.body.loc.body_dir));
}

void update_world_turn(double dir)
{
  world.body.loc.body_dir = normalize_angle(world.body.loc.body_dir+dir/(1.0+INERTIA_MOMENT*speed(world.body.loc)));
}

void check_counters(void)
{
  bool chst = true;

  if (sensor.lost_message)
    world.body.plcount = world.counter;

  sensor.lost_message = false;
  
  if (world.counter.turn != world.body.plcount.turn) chst = false;
  if (world.counter.dash != world.body.plcount.dash) chst = false;
  if (world.counter.kick != world.body.plcount.kick) chst = false;
  if (world.counter.move != world.body.plcount.move) chst = false;

  if (!chst) {
    sensor.lost_message = true;
    world.body.plcount = world.counter;
    world.body.loc = world.tmate[player.unum];
    logmsg("Message lost : Counters mismatch!!");
    report_error("Message lost - Slow network or slow machine");
  };
}

void update_offside_info(void)
{
  int m1, m2, i;
  
  m1 = 1;
  for(i=1;i<=N_PLAYER_PER_TEAM;i++) if (world.tmate[i].correctness>MIN_USEFUL_CRR)
    if (world.tmate[m1].x>world.tmate[i].x) m1=i;
  m2 = (m1==1)?2:1;
  for(i=1;i<=N_PLAYER_PER_TEAM;i++) if (world.tmate[i].correctness>MIN_USEFUL_CRR)
    if (i!=m1)
      if (world.tmate[m2].x>world.tmate[i].x) m2=i;
  world.offside.our.line = world.tmate[m2].x;
  world.offside.our.unum = m2;
  world.offside.our.offpos = false;
  if (world.offside.our.line<world.ball.x)
    for(i=1;i<=N_PLAYER_PER_TEAM;i++) if (world.opp[i].correctness>MIN_USEFUL_CRR)
      if (world.opp[i].x<world.offside.our.line) world.offside.our.offpos = true;

  m1 = 1;
  for(i=1;i<=N_PLAYER_PER_TEAM;i++) if (world.opp[i].correctness>MIN_USEFUL_CRR)
    if (world.opp[m1].x<world.opp[i].x) m1=i;
  m2 = (m1==1)?2:1;
  for(i=1;i<=N_PLAYER_PER_TEAM;i++) if (world.opp[i].correctness>MIN_USEFUL_CRR)
    if (i!=m1)
      if (world.opp[m2].x<world.opp[i].x) m2=i;
  world.offside.opp.line = world.opp[m2].x;
  world.offside.opp.unum = m2;
  world.offside.opp.offpos = false;
  if (world.offside.opp.line>world.ball.x)
    for(i=1;i<=N_PLAYER_PER_TEAM;i++) if (world.tmate[i].correctness>MIN_USEFUL_CRR)
      if (world.tmate[i].x>world.offside.opp.line) world.offside.opp.offpos = true;

  world.offside.inoff = (world.offside.opp.line<world.body.loc.x);
}

void update_sent_commands(void)
{
  sensor.just_sent_command.type = NO_COMMAND;
  sensor.just_sent_command.clock = world.time.clock;
  sensor.just_sent_command.ack = false;
  sensor.just_sent_command.neck_moment = 0;


  if (sensor.lost_message) {
    sensor.just_sent_command = sensor.last_sent_command;
    sensor.last_sent_command.type = NO_COMMAND;
  }

  switch(sensor.last_sent_command.type) {
    case NO_COMMAND:
    break;
    case KICK:
      update_world_kick(sensor.last_sent_command.par1,sensor.last_sent_command.par2);
    break;
    case DASH:
      update_world_dash(sensor.last_sent_command.par1);
    break;
    case TURN:
      update_world_turn(sensor.last_sent_command.par1);
    break;
  }
  world.body.head_angle += sensor.last_sent_command.neck_moment;

}

void update_world_objects(void)
{
  int i;
  double dir, t1, t2;
/*  
  if ((world.body.ball_near_sensor) && (sensor.last_sent_command.type!=KICK)) {
    sprintf(debugbuf," Pooya: v1=%.2f,%.2f ",world.ball.vx*BALL_DECAY,world.ball.vy*BALL_DECAY);
    logmsg(debugbuf);
    world.ball.vx = world.ball.x-world.ball_last_cycle.x;
    world.ball.vy = world.ball.y-world.ball_last_cycle.y;
    world.ball.vx *= BALL_DECAY;
    world.ball.vy *= BALL_DECAY;
    sprintf(debugbuf, " v2=%.2f,%.2f \n",world.ball.vx,world.ball.vy);
    logmsg(debugbuf);
  };

  world.ball_last_cycle = world.ball;
  world.body.ball_near_sensor = false;
*/
  world.body.loc.correctness -= CORRECTNESS_DECAY;
  if (world.body.loc.correctness<0) world.body.loc.correctness=0;
  t1 = estimate_mypos_x(1);
  t2 = estimate_mypos_y(1);
  world.body.loc.x = t1;
  world.body.loc.y = t2;
  world.body.loc.vx *= PLAYER_DECAY;
  world.body.loc.vy *= PLAYER_DECAY;

  for(i=1;i<=N_PLAYER_PER_TEAM;i++) {
    world.opp[i].correctness -= CORRECTNESS_DECAY;
    if (world.opp[i].correctness<0) world.opp[i].correctness=0;
    t1 = estimate_player_x(world.opp[i],1);
    t2 = estimate_player_y(world.opp[i],1);
    world.opp[i].x = t1;
    world.opp[i].y = t2;
    world.opp[i].vx *= PLAYER_DECAY;
    world.opp[i].vy *= PLAYER_DECAY;

    world.tmate[i].correctness -= CORRECTNESS_DECAY;
    if (world.tmate[i].correctness<0) world.tmate[i].correctness=0;
    t1 = estimate_player_x(world.tmate[i],1);
    t2 = estimate_player_y(world.tmate[i],1);
    world.tmate[i].x = t1;
    world.tmate[i].y = t2;
    world.tmate[i].vx *= PLAYER_DECAY;
    world.tmate[i].vy *= PLAYER_DECAY;
  }

  world.ball.correctness -= CORRECTNESS_DECAY;
  t1 = estimate_ball_x(1);
  t2 = estimate_ball_y(1);
  world.ball.x = t1;
  world.ball.y = t2;
  world.ball.vx *= BALL_DECAY;
  world.ball.vy *= BALL_DECAY;
  if (world.ball.correctness<0) world.ball.correctness=0;
 
 // Collisions
  if (dist_2me(world.ball)<PLAYER_SIZE+BALL_SIZE) {
    world.body.loc.vx *= COLLISION_VELOCITY;
    world.body.loc.vy *= COLLISION_VELOCITY;
    world.ball.vx *= COLLISION_VELOCITY;
    world.ball.vy *= COLLISION_VELOCITY;
    dir = Deg2Rad(dir_from_me(world.ball));
    world.ball.x = world.body.loc.x-(PLAYER_SIZE+BALL_SIZE)*cos(dir);
    world.ball.y = world.body.loc.y-(PLAYER_SIZE+BALL_SIZE)*sin(dir);
  }
}

void update_ballcontrol(void)
{
  if (can_kickball()) {
    world.control_ball.incontrol = true;
    world.control_ball.last_kickable_cycle = world.time.clock;
    world.control_ball.getball = false;
    sensor.comm.passdest = player.unum;
  } else {
    if (dist_2me(world.ball)>BALL_CONTROL_RANGE) world.control_ball.incontrol = false;
    if (world.time.clock-world.control_ball.last_kickable_cycle>BALL_CONTROL_TURNS) world.control_ball.incontrol = false;
    if ((speed(world.ball)>BALL_CONTROL_SPEED)&&(dist_2me(estimate_ball_x(1),estimate_ball_y(1)))>KICKABLE_AREA) world.control_ball.incontrol = false;
  }
//  if (!world.control_ball.incontrol) world.body.ball_inhand = false;

  if (dist_2me(world.ball)>CATCHABLE_AREA_L) world.body.ball_inhand = false;
}

void update_newturn(void)
{
  world.body.head_angle = sensor.body.head_angle;
  world.body.speed.val = sensor.body.speed_val;
  world.body.speed.ang = normalize_angle(sensor.body.speed_dir+world.body.loc.body_dir+world.body.head_angle);
  world.body.loc.vx = sensor.body.speed_val*cos(Deg2Rad(world.body.speed.ang));
  world.body.loc.vy = sensor.body.speed_val*sin(Deg2Rad(world.body.speed.ang));
}

void update_skills_bytime(void)
{
  pass.update_bytime();
  dribble.update_bytime();
  double_kick.update_bytime();
  shoot2goal.update_bytime();
  get_2ball.update_bytime();
  goto_playing_post.update_bytime();  
}

void update_world_bytime(void)
{
  check_counters();
  update_sent_commands();
  update_world_objects();
  update_offside_info();
  update_ballcontrol();
  if (Time2Say()) say_world_model();
  update_newturn();
}
